#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import json

from directord import components
from directord import utils


class Component(components.ComponentBase):
    def __init__(self):
        """Initialize the component cache class.

        This component is not cacheable.
        """

        super().__init__(desc="Process cachefile commands")
        self.cacheable = False
        self.requires_lock = True

    def args(self):
        """Set default arguments for a component."""

        super().args()
        self.parser.add_argument(
            "--no-wait",
            action="store_true",
            help=(
                "Disable waiting for the queried key to be present in the"
                " local cache."
            ),
        )
        self.parser.add_argument(
            "query",
            help=(
                "Scan the environment for a given cached argument and"
                " store the resultant on the target. The resultant is"
                " set in dictionary format: `{'client-id': ...}`"
            ),
        )

    def server(self, exec_array, data, arg_vars):
        """Return data from formatted transfer action.

        :param exec_array: Input array from action
        :type exec_array: List
        :param data: Formatted data hash
        :type data: Dictionary
        :param arg_vars: Pre-Formatted arguments
        :type arg_vars: Dictionary
        :returns: Dictionary
        """

        super().server(exec_array=exec_array, data=data, arg_vars=arg_vars)
        data["query"] = self.known_args.query
        data["no_wait"] = self.known_args.no_wait
        return data

    def client(self, cache, job):
        """Run query command operation.

        :param cache: Caching object used to template items within a command.
        :type cache: Object
        :param job: Information containing the original job specification.
        :type job: Dictionary
        :returns: tuple
        """

        self.log.debug("client(): job: %s, cache: %s", job, cache)
        args = cache.get("args")
        if args:
            query = args.get(job["query"])
        else:
            query = None

        self.block_on_tasks = list()
        arg_job = job.copy()
        query_item = arg_job.pop("query")
        targets = arg_job.get("targets", list())
        arg_job.pop("parent_sha3_224", None)
        arg_job.pop("parent_id", None)
        arg_job.pop("job_sha3_224", None)
        arg_job.pop("job_id", None)
        arg_job["skip_cache"] = True
        arg_job["extend_args"] = True
        arg_job["verb"] = "ARG"
        arg_job["args"] = {
            "query": {self.driver.identity: {query_item: query}}
        }
        arg_job["parent_async_bypass"] = True
        arg_job["job_id"] = utils.get_uuid()
        arg_job["job_sha3_224"] = utils.object_sha3_224(obj=arg_job)
        arg_job["parent_id"] = utils.get_uuid()
        arg_job["parent_sha3_224"] = utils.object_sha3_224(obj=arg_job)
        self.block_on_tasks.append(arg_job)
        if self.driver.identity in targets:
            if not job.get("no_wait"):
                wait_job = dict(
                    skip_cache=True,
                    verb="QUERY_WAIT",
                    item=query_item,
                    query_timeout=600,
                    parent_async_bypass=True,
                    targets=targets,
                    identity=list(),
                )
                wait_job["job_id"] = utils.get_uuid()
                wait_job["job_sha3_224"] = utils.object_sha3_224(obj=wait_job)
                wait_job["parent_id"] = arg_job["parent_id"]
                wait_job["parent_sha3_224"] = arg_job["parent_sha3_224"]
                self.block_on_tasks.append(wait_job)

        return json.dumps(query), None, True, None



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import flask

from flask import request
from flask.views import MethodView


APP = flask.Flask(__name__, instance_relative_config=True)
JOBS = list()
NODES = list()


class DirectordUI(MethodView):
    """Implement a method view for the Directord UI."""

    @staticmethod
    def get():
        """Return a get response from a template."""

        response = flask.make_response(
            flask.render_template(
                "index.html",
                remote_url=request.base_url,
                count=len(JOBS.keys()),
                jobs=JOBS.items(),
                nodes=NODES.items(),
            )
        )
        response.headers = {
            "X-Frame-Options": "SAMEORIGIN",
            "Cache-Control": "public, max-age=120",
        }
        return response


class UI:
    """The Directord UI execution class."""

    def __init__(self, args, jobs, nodes):
        """Initialize the Directord UI class.

        :param args: Arguments parsed by argparse.
        :type args: Object
        :params jobs: A manager object for all jobs.
        :type jobs: Dictionary
        :params nodes: A manager object for all nodes.
        :type nodes: Dictionary
        """

        global JOBS
        JOBS = jobs
        global NODES
        NODES = nodes
        self.args = args

    def start_app(self):
        """Start the application in production mode.

        Starts the flask application using the embedded server.
        """

        APP.add_url_rule("/", view_func=DirectordUI.as_view("index"))
        if self.args.bind_address == "*":
            host = "0.0.0.0"
        else:
            host = self.args.bind_address
        APP.run(port=self.args.ui_port, host=host)



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import json
import time

import redis


class BaseDocument:
    """Create a document store object."""

    def __init__(self, url, database=0):
        """Initialize the redis datastore.

        :param url: Connection string to the redis deployment.
        :type url: String
        :param database: Keyspace used for Redis
        :type database: Integer
        """

        self.datastore = redis.Redis.from_url(url=url, db=database)

    def __getitem__(self, key):
        """Return the value of a given key.

        Objects stored in the keyspace are JSON, the getter will
        decode and load the JSON into a standard python object.

        :param key: Named object.
        :type key: Object
        :returns: Object
        """

        value = self.datastore.get(key)
        if value:
            value = value.decode()
            try:
                value = json.loads(value)
            except Exception:
                return value
            else:
                return value

    def __setitem__(self, key, value):
        """Set an item in the datastore.

        objects are serialized JSON.

        :param key: Named object to set.
        :type key: Object
        :param value: Object to set.
        :type value: Object
        """

        self.datastore.set(key, json.dumps(value))

    def __delitem__(self, key):
        """Delete an item from the datastore.

        :param key: Named object.
        :type key: Object
        """

        self.datastore.delete(key)

    def items(self):
        """Yield a tuple for key and value."""

        for item in self.datastore.keys("*"):
            value = self.__getitem__(item)
            if value:
                yield item.decode(), value

    def keys(self):
        """Return an array of all keys.

        :returns: List
        """

        return self.datastore.keys("*")

    def empty(self):
        """Empty all items from the datastore.

        Because a Manager Dict is a proxy object we don't want to replace the
        object we want to empty it keeping the original proxy intact. This
        method will pop all items from the object.
        """

        self.datastore.flushdb()

    def pop(self, key):
        """Delete the value of a given key.

        :param key: Named object.
        :type key: Object
        """

        self.__delitem__(key)

    def prune(self):
        """Prune items that have a time based expiry."""

        for item in self.datastore.keys("*"):
            value = self.__getitem__(item)
            try:
                if time.time() >= value["time"]:
                    self.datastore.delete(item)
            except (KeyError, TypeError):
                pass

        return len(self.datastore.keys("*"))

    def get(self, key):
        """Return the value of a given key.

        :param key: Named object.
        :type key: Object
        :returns: Object
        """

        return self.__getitem__(key)

    def set(self, key, value):
        """Set key and value if key doesn't already exist.

        :param key: Named object to set.
        :type key: Object
        :param value: Object to set.
        :type value: Object
        :returns: Object
        """

        item = self.__getitem__(key)
        if item:
            return item

        self.__setitem__(key, value)
        return value



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import logging
import os
import directord


class Interface(directord.Processor):
    """The Interface class.

    This class defines everything required to connect to or from a given
    server.
    """

    def __init__(self, args):
        """Initialize the interface class.

        :param args: Arguments parsed by argparse.
        :type args: Object
        """

        super(Interface, self).__init__()

        self.args = args

        # Set log handlers to debug when enabled.
        if self.args.debug:
            self.log.setLevel(logging.DEBUG)
            for handler in self.log.handlers:
                handler.setLevel(logging.DEBUG)

        mode = getattr(self.args, "mode", None)
        if mode == "client":
            self.bind_address = self.args.server_address
        elif mode == "server":
            self.bind_address = self.args.bind_address
        else:
            self.bind_address = "*"

        self.proto = "tcp"
        self.connection_string = "{proto}://{addr}".format(
            proto=self.proto, addr=self.bind_address
        )

        try:
            self.heartbeat_interval = self.args.heartbeat_interval
        except AttributeError:
            self.heartbeat_interval = 1

        self.base_dir = "/etc/directord"
        self.public_keys_dir = os.path.join(self.base_dir, "public_keys")
        self.secret_keys_dir = os.path.join(self.base_dir, "private_keys")
        self.keys_exist = os.path.exists(
            self.public_keys_dir
        ) and os.path.exists(self.secret_keys_dir)

        self.log.debug("Loading messaging driver")

        try:
            _driver = directord.plugin_import(
                plugin=".drivers.{}".format(self.args.driver)
            )
        except Exception as e:
            raise SystemExit(
                "Driver was not able to be loaded: {}".format(str(e))
            )
        else:
            self.driver = _driver.Driver(
                args=self.args,
                encrypted_traffic_data={
                    "enabled": self.keys_exist,
                    "public_keys_dir": self.public_keys_dir,
                    "secret_keys_dir": self.secret_keys_dir,
                },
                connection_string=self.connection_string,
            )



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import hashlib
import json
import os
import socket
import time
import uuid

import tabulate
import yaml

from ssh import options
from ssh.session import Session
from ssh import key as ssh_key

from directord import logger


def dump_yaml(file_path, data):
    """Dump data to a file.

    :param file_path: File path to dump data to
    :type file_path: String
    :param data: Dictionary|List data to dump
    :type data: Dictionary|List
    """

    with open(os.path.abspath(os.path.expanduser(file_path)), "w") as f:
        yaml.safe_dump(data, f, default_flow_style=False)

    return file_path


def merge_dict(base, new, extend=True):
    """Recursively merge new into base.

    :param base: Base dictionary to load items into
    :type base: Dictionary
    :param new: New dictionary to merge items from
    :type new: Dictionary
    :param extend: Boolean option to enable or disable extending
                   iterable arrays.
    :type extend: Boolean
    :returns: Dictionary
    """

    if isinstance(new, dict):
        for key, value in new.items():
            if key not in base:
                base[key] = value
            elif extend and isinstance(value, dict):
                base[key] = merge_dict(
                    base=base.get(key, {}), new=value, extend=extend
                )
            elif extend and isinstance(value, list):
                base[key].extend(value)
            elif extend and isinstance(value, (tuple, set)):
                if isinstance(base.get(key), tuple):
                    base[key] += tuple(value)
                elif isinstance(base.get(key), set):
                    base[key].update(value)
            else:
                base[key] = new[key]
    elif isinstance(new, list):
        if extend:
            base.extend(new)
        else:
            base = new

    return base


class ClientStatus:
    """Context manager for transmitting client status."""

    def __init__(self, socket, job_id, command, ctx):
        """Initialize the UNIX socket connect context manager."""

        self.ctx = ctx
        self.job_id = job_id
        self.command = command
        self.job_state = ctx.driver.nullbyte
        self.info = ctx.driver.nullbyte
        self.socket = socket
        self.data = None
        self.stderr = ctx.driver.nullbyte
        self.stdout = ctx.driver.nullbyte

    def __enter__(self):
        """Upon enter, return the context manager object for future updates.

        :returns: Object
        """

        return self

    def __exit__(self, *args, **kwargs):
        """Upon exit, send a final status message."""

        try:
            self.stderr = self.stderr.encode()
        except AttributeError:
            pass

        try:
            self.stdout = self.stdout.encode()
        except AttributeError:
            pass

        try:
            self.info = self.info.encode()
        except AttributeError:
            pass

        self.ctx.driver.socket_send(
            socket=self.socket,
            msg_id=self.job_id,
            control=self.job_state,
            command=self.command,
            data=self.data,
            info=self.info,
            stderr=self.stderr,
            stdout=self.stdout,
        )


class SSHConnect:
    """Context manager to remotely connect to servers using libssh.

    The connection manager requires an SSH key to be defined, and exist,
    however, upon enter the system will use the SSH agent is defined.
    """

    def __init__(self, host, username, port, key_file=None, debug=False):
        """Initialize the connection manager.

        :param host: IP or Domain to connect to.
        :type host: String
        :param username: Username for the connection.
        :type username: String
        :param port: Port number used to connect to the remote server.
        :type port: Int
        :param key_file: SSH key file used to connect.
        :type key_file: String
        :param debug: Enable or disable debug mode
        :type debug: Boolean
        """

        self.log = logger.getLogger(name="directord", debug_logging=debug)
        self.key_file = key_file
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.connect((host, port))

        self.session = Session()
        self.session.options_set(options.HOST, host)
        self.session.options_set(options.USER, username)
        self.session.options_set_port(port)
        self.session.set_socket(self.sock)
        self.session.connect()

        self.log.debug(
            "Handshake with [ %s ] on port [ %s ] complete.", host, port
        )

        self.channels = dict()
        self.host = host
        self.username = username
        self.key_file = key_file

    def _userauth_publickey_fromfile(self, key_file):
        """Import a private key file.

        :param key_file: Fully qualified path to an ssh key file.
        :type key_file: String
        """

        key = ssh_key.import_privkey_file(key_file)
        self.session.userauth_publickey(key)

    def set_auth(self):
        """Set the ssh session auth."""

        if self.key_file:
            self._userauth_publickey_fromfile(key_file=self.key_file)
            self.log.debug("Key file [ %s ] added", self.key_file)
        else:
            try:
                self.session.userauth_agent(self.username)
                self.log.debug("User agent based authentication enabled")
            except Exception as e:
                self.log.warning(
                    "SSH Agent connection has failed: %s."
                    " Attempting to connect with the user's implicit ssh key.",
                    str(e),
                )
                home = os.path.abspath(os.path.expanduser("~"))
                default_keyfile = os.path.join(home, ".ssh/id_rsa")
                if os.path.exists(default_keyfile):
                    self._userauth_publickey_fromfile(key_file=default_keyfile)
                    self.log.debug(
                        "Implicit key file [ %s ] added", self.key_file
                    )
                else:
                    self.log.critical(
                        "No implicit key found [ %s ]. Setup user-agent"
                        " authentication or use the --key-file"
                        " argument to explicitly set the required"
                        " ssh key.",
                        default_keyfile,
                    )
                    raise SystemExit("Authentication failure")

    def __enter__(self):
        """Connect to the remote node and return the ssh and session objects.

        :returns: Tuple
        """

        self.set_auth()
        return self

    def __exit__(self, *args, **kwargs):
        """Upon exit, close the ssh connection."""

        for key, value in self.channels.items():
            if hasattr(value, "close"):
                value.close()
                self.log.debug("%s channel is closed.", key)

        self.session.disconnect()
        self.log.debug("SSH session is closed.")


def file_sha3_224(file_path, chunk_size=10240):
    """Return the SHA3_224 sum of a given file.

    Default chunk size: 10K.

    :param file_path: File path
    :type file_path: String
    :param chunk_size: Set the read chunk size.
    :type chunk_size: Integer
    :returns: String
    """

    sha3_224 = hashlib.sha3_224()
    if os.path.exists(file_path):
        with open(file_path, "rb") as f:
            while True:
                data = f.read(chunk_size)
                if not data:
                    break
                else:
                    sha3_224.update(data)

        return sha3_224.hexdigest()


def object_sha3_224(obj):
    """Return the SHA3_224 sum of a given object.

    The object used for generating a SHA3_224 must be JSON compatible.

    :param file_path: File path
    :type file_path: String
    :returns: String
    """

    return hashlib.sha3_224(json.dumps(obj).encode()).hexdigest()


def get_uuid():
    """Return a new UUID in String format.

    :returns: String
    """

    return str(uuid.uuid4())


def print_tabulated_data(data, headers):
    """Print data in tabulated form.

    :param data: Organized data
    :type data: List
    :param headers: List of headers
    :type headers: List
    """

    print(
        tabulate.tabulate(
            data,
            headers=headers,
            disable_numparse=True,
        )
    )


def return_poller_interval(poller_time, poller_interval, log=None):
    """Return a new poller interval time.

    Review the poller time vs the current time, and if the poller is outside
    our expected margins return a new poller time to cool down the poller
    processes.

    :returns: Integer
    """

    current_time = time.time()
    if current_time >= poller_time + 64:
        if poller_interval != 2048:
            if log:
                log.debug("Directord entering idle state.")
        poller_interval = 2048
    elif current_time >= poller_time + 32:
        if poller_interval != 1024:
            if log:
                log.debug("Directord ramping down.")
        poller_interval = 1024

    return poller_interval



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from directord import components


try:
    import selinux

    AVAILABLE_SELINUX = True
except ImportError:
    AVAILABLE_SELINUX = False


try:
    import seobject

    AVAILABLE_SEOBJECT = True
except ImportError:
    AVAILABLE_SEOBJECT = False


class Component(components.ComponentBase):
    def __init__(self):
        super().__init__(desc="Process secontext commands")
        self.file_type_str = {
            "a": "all files",
            "b": "block device",
            "c": "character device",
            "d": "directory",
            "f": "regular file",
            "l": "symbolic link",
            "p": "named pipe",
            "s": "socket",
        }

    def semanage_exists(self, secontext, target, ftype):
        """Get the SELinux file context mapping definition from policy.

        :secontext: Object
        :target: String
        :ftype: String
        :returns: Tuple
        """

        record = (target, self.file_type_str[ftype])
        records = secontext.get_all()
        orig_seuser, _, _, orig_selevel = records[record]
        try:
            return True, orig_seuser, orig_selevel
        except KeyError:
            return False, None, None

    def args(self):
        """Set default arguments for a component."""

        super().args()
        self.parser.add_argument(
            "--ftype",
            help="The contexts type. %(default)s",
            choices=list(self.file_type_str.keys()),
        )
        self.parser.add_argument(
            "--reload",
            help="Reload policy after commit",
            default=False,
            type=bool,
        )
        self.parser.add_argument(
            "--selevel",
            help="Selinux level",
            default=str,
        )
        self.parser.add_argument(
            "--setype", help="Selinux type.", default=str, required=True
        )
        self.parser.add_argument(
            "--seuser",
            help="Selinux user.",
            default=str,
        )
        self.parser.add_argument(
            "target",
            help="Add or modify the selinux context for a given target.",
        )

    def server(self, exec_array, data, arg_vars):
        """Return data from formatted cacheevict action.

        :param exec_array: Input array from action
        :type exec_array: List
        :param data: Formatted data hash
        :type data: Dictionary
        :param arg_vars: Pre-Formatted arguments
        :type arg_vars: Dictionary
        :returns: Dictionary
        """

        super().server(exec_array=exec_array, data=data, arg_vars=arg_vars)
        if self.known_args.ftype:
            data["ftype"] = self.known_args.ftype

        if self.known_args.setype:
            data["setype"] = self.known_args.setype

        if self.known_args.seuser:
            data["seuser"] = self.known_args.seuser

        if self.known_args.selevel:
            data["selevel"] = self.known_args.selevel

        if self.known_args.target:
            data["target"] = self.known_args.target

        data["reload"] = self.known_args.reload

        return data

    def client(self, cache, job):
        """Run cache echo command operation.

        :param cache: Caching object used to template items within a command.
        :type cache: Object
        :param job: Information containing the original job specification.
        :type job: Dictionary
        :returns: tuple
        """

        if not selinux.is_selinux_enabled():
            return (None, "SELinux is not enabled.", True, None)

        if not AVAILABLE_SELINUX:
            return (
                None,
                "The required selinux library is not installed",
                False,
                None,
            )

        if not AVAILABLE_SEOBJECT:
            return (
                None,
                "The required seobject library is not installed",
                False,
                None,
            )

        target = job["target"]
        ftype = job.get("ftype")
        setype = job["setype"]
        seuser = job.get("seuser")
        selevel = job.get("selevel")

        secontext = seobject.fcontextRecords("")
        secontext.set_reload(job["reload"])

        (
            existing,
            orig_seuser,
            orig_selevel,
        ) = self.semanage_exists(secontext, target, ftype)

        if existing:
            if not seuser:
                seuser = orig_seuser

            if not selevel:
                selevel = orig_selevel

            secontext.modify(target, setype, ftype, selevel, seuser)
        else:
            if not seuser:
                seuser = "system_u"

            if not selevel:
                selevel = "s0"

            secontext.add(target, setype, ftype, selevel, seuser)

        return job["target"], None, True, None



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import decimal
import grp
import json
import os
import socket
import time
import urllib.parse as urlparse

import directord

from directord import interface
from directord import utils


class Server(interface.Interface):
    """Directord server class."""

    def __init__(self, args):
        """Initialize the Server class.

        Sets up the server object.

        :param args: Arguments parsed by argparse.
        :type args: Object
        """

        super(Server, self).__init__(args=args)
        self.job_queue = self.get_queue()
        self.send_queue = self.get_queue()
        datastore = getattr(self.args, "datastore", None)
        if not datastore or datastore == "memory":
            self.log.info("Connecting to internal datastore")
            directord.plugin_import(plugin=".datastores.internal")
            manager = self.get_manager()
            self.workers = manager.document()
            self.return_jobs = manager.document()
        else:
            url = urlparse.urlparse(datastore)
            if url.scheme in ["file"]:
                disc = directord.plugin_import(plugin=".datastores.disc")
                self.log.debug("Disc base document store initialized")
                path = os.path.abspath(os.path.expanduser(url.path))
                # Ensure that the cache path exists before executing.
                os.makedirs(path, exist_ok=True)
                self.workers = disc.BaseDocument(
                    url=os.path.join(path, "workers")
                )
                self.workers.empty()
                self.return_jobs = disc.BaseDocument(
                    url=os.path.join(path, "jobs")
                )
            if url.scheme in ["redis", "rediss"]:
                self.log.info("Connecting to redis datastore")
                try:
                    db = int(url.path.lstrip("/"))
                except ValueError:
                    db = 0
                self.log.debug("Redis keyspace base is %s", db)
                redis = directord.plugin_import(plugin=".datastores.redis")

                self.workers = redis.BaseDocument(
                    url=url._replace(path="").geturl(), database=(db + 1)
                )
                self.return_jobs = redis.BaseDocument(
                    url=url._replace(path="").geturl(), database=(db + 2)
                )

    def _set_job_status(
        self,
        job_status,
        job_id,
        identity,
        job_output,
        job_stdout=None,
        job_stderr=None,
        execution_time=None,
        return_timestamp=None,
        component_exec_timestamp=None,
        recv_time=None,
    ):
        """Set job status.

        This will update the object for job tracking, allowing the
        user to know what happened within the environment.

        :param job_status: ASCII Control Character
        :type job_status: Bytes
        :param job_id: UUID for job
        :type job_id: String
        :param identity: Node name
        :type identity: String
        :param job_output: Job output information
        :type job_output: String
        :param job_stdout: Job output information
        :type job_stdout: String
        :param job_stderr: Job error information
        :type job_stderr: String
        :param execution_time: Time the task took to execute
        :type execution_time: Float
        :param return_timestamp: Timestamp from the job return.
        :type return_timestamp: String
        :param component_exec_timestamp: Timestamp from the component return.
        :type component_exec_timestamp: String
        :param recv_time: Time a task return was received.
        :type recv_tim: Float
        """

        def _set_time():
            _createtime = job_metadata.get("_createtime")
            if not _createtime:
                _createtime = job_metadata["_createtime"] = time.time()

            if isinstance(recv_time, (int, float)):
                job_metadata["_roundtripltime"][identity] = (
                    float(recv_time) - _createtime
                )
                job_metadata["ROUNDTRIP_TIME"] = "{:.8f}".format(
                    decimal.Decimal(
                        sum(job_metadata["_roundtripltime"].values())
                        / len(job_metadata["_roundtripltime"].keys())
                    )
                )

            if isinstance(execution_time, (int, float)):
                job_metadata["_executiontime"][identity] = float(
                    execution_time
                )
                job_metadata["EXECUTION_TIME"] = "{:.8f}".format(
                    decimal.Decimal(
                        sum(job_metadata["_executiontime"].values())
                        / len(job_metadata["_executiontime"].keys())
                    )
                )

        job_metadata = self.return_jobs.get(job_id)
        if not job_metadata:
            return

        if job_output and job_output is not self.driver.nullbyte.decode():
            job_metadata["INFO"][identity] = job_output

        if job_stdout and job_stdout is not self.driver.nullbyte.decode():
            job_metadata["STDOUT"][identity] = job_stdout

        if job_stderr and job_stderr is not self.driver.nullbyte.decode():
            job_metadata["STDERR"][identity] = job_stderr

        self.log.debug("current job [ %s ] state [ %s ]", job_id, job_status)
        job_metadata["_processing"][identity] = job_status.decode()
        if job_status == self.driver.job_ack:
            self.log.debug("%s received job %s", identity, job_id)
        elif job_status == self.driver.job_processing:
            self.log.debug("%s is processing %s", identity, job_id)
        elif job_status == self.driver.job_end:
            _set_time()
            self.log.debug("%s finished processing %s", identity, job_id)
            if "SUCCESS" in job_metadata:
                job_metadata["SUCCESS"].append(identity)
            else:
                job_metadata["SUCCESS"] = [identity]
            job_metadata["_processing"][
                identity
            ] = self.driver.job_end.decode()
        elif job_status == self.driver.job_failed:
            _set_time()
            self.log.debug("%s failed %s", identity, job_id)
            if "FAILED" in job_metadata:
                job_metadata["FAILED"].append(identity)
            else:
                job_metadata["FAILED"] = [identity]

        if return_timestamp:
            job_metadata["RETURN_TIMESTAMP"] = return_timestamp

        if component_exec_timestamp:
            job_metadata["COMPONENT_TIMESTAMP"] = component_exec_timestamp

        for process in job_metadata["_processing"].values():
            if process == self.driver.job_processing.decode():
                job_metadata[
                    "PROCESSING"
                ] = self.driver.job_processing.decode()
                break
        else:
            job_metadata["PROCESSING"] = self.driver.job_end.decode()

        job_metadata["_lasttime"] = time.time()
        self.return_jobs[job_id] = job_metadata

    def create_return_jobs(self, task, job_item, targets):
        return self.return_jobs.set(
            task,
            {
                "ACCEPTED": True,
                "INFO": dict(),
                "STDOUT": dict(),
                "STDERR": dict(),
                "_nodes": [i.decode() for i in targets],
                "VERB": job_item["verb"],
                "TRANSFERS": list(),
                "JOB_SHA3_224": job_item["job_sha3_224"],
                "JOB_DEFINITION": job_item,
                "PARENT_JOB_ID": job_item.get("parent_id"),
                "_createtime": time.time(),
                "_executiontime": dict(),
                "_roundtripltime": dict(),
                "_processing": dict(),
            },
        )

    def run_job(self, sentinel=False):
        """Run a job interaction

        As the job loop executes it will interrogate the job item as returned
        from the queue. If the item contains a "targets" definition the
        job loop will only send the message to the given targets, assuming the
        target is known within the workers object, otherwise all targets will
        receive the message. If a defined target is not found within the
        workers object no job will be executed.

        :param sentinel: Breaks the loop
        :type sentinel: Boolean
        :returns: Tuple
        """

        self.log.info("Starting run process.")
        while not self.job_queue.empty():
            try:
                job_item = self.job_queue.get(timeout=60)
            except Exception:
                break
            else:
                self.log.debug("Job item received [ %s ]", job_item)
                restrict_sha3_224 = job_item.get("restrict")
                if restrict_sha3_224:
                    if job_item["job_sha3_224"] not in restrict_sha3_224:
                        self.log.debug(
                            "Job restriction %s is unknown.", restrict_sha3_224
                        )
                        if sentinel:
                            break
                        else:
                            continue

                targets = list()
                self.log.debug("Processing targets.")
                for target in (
                    job_item.pop("targets", None) or self.workers.keys()
                ):
                    try:
                        target = target.encode()
                    except AttributeError:
                        pass

                    if target in self.workers.keys():
                        self.log.debug("Target identified [ %s ].", target)
                        targets.append(target)
                    else:
                        self.log.critical(
                            "Target [ %s ] is unknown. Check the name againt"
                            " the available targets",
                            target,
                        )
                        if not self.return_jobs.get(job_item["job_id"]):
                            self.create_return_jobs(
                                task=job_item["job_id"],
                                job_item=job_item,
                                targets=targets,
                            )
                        self._set_job_status(
                            job_status=self.driver.job_failed,
                            job_id=job_item["job_id"],
                            identity=target.decode(),
                            job_output=(
                                "Target unknown. Available targets {}".format(
                                    list(self.workers.keys())
                                )
                            ),
                            recv_time=time.time(),
                        )

                if not targets:
                    self.log.error("No known targets defined.")
                    continue

                self.log.debug("All targets %s", targets)
                if job_item["verb"] == "QUERY":
                    self.log.debug("Query mode enabled.")
                    # NOTE(cloudnull): QUERY runs across the cluster. The
                    #                  callback tasks are scoped to only
                    #                  the nodes defined within the job
                    #                  execution.
                    job_item["targets"] = [i.decode() for i in targets]
                    targets = self.workers.keys()
                elif job_item.get("run_once", False):
                    self.log.debug("Run once enabled.")
                    targets = [targets[0]]
                    job_item["targets"] = [targets[0].decode()]

                job_id = job_item.get("job_id", utils.get_uuid())
                job_info = self.create_return_jobs(
                    task=job_id, job_item=job_item, targets=targets
                )
                self.log.debug("Processing job [ %s ]", job_item)
                for identity in targets:
                    if job_item["verb"] in ["ADD", "COPY"]:
                        for file_path in job_item["from"]:
                            job_item["file_sha3_224"] = utils.file_sha3_224(
                                file_path=file_path
                            )
                            if job_item["to"].endswith(os.sep):
                                job_item["file_to"] = os.path.join(
                                    job_item["to"],
                                    os.path.basename(file_path),
                                )
                            else:
                                job_item["file_to"] = job_item["to"]

                            if (
                                job_item["file_to"]
                                not in job_info["TRANSFERS"]
                            ):
                                job_info["TRANSFERS"].append(
                                    job_item["file_to"]
                                )

                            self.log.debug(
                                "Queueing file transfer job [ %s ] for"
                                " file_path [ %s ] to identity [ %s ]",
                                job_item["job_id"],
                                file_path,
                                identity.decode(),
                            )
                            self.send_queue.put(
                                dict(
                                    identity=identity,
                                    command=job_item["verb"].encode(),
                                    data=job_item,
                                    info=file_path.encode(),
                                )
                            )
                    else:
                        self.log.debug(
                            "Queuing job [ %s ] for identity [ %s ]",
                            job_item["job_id"],
                            identity.decode(),
                        )
                        self.send_queue.put(
                            dict(
                                identity=identity,
                                command=job_item["verb"].encode(),
                                data=job_item,
                            )
                        )
                else:
                    self.return_jobs[job_id] = job_info

            if sentinel:
                break

    def run_backend(self, sentinel=False):
        """Execute the transfer loop.

        Directord's interaction executor will slow down the poll interval
        when no work is present. This means Directord will ramp-up resource
        utilization when required and become virtually idle when there's
        nothing to do.

        * Initial poll interval is 1024, maxing out at 2048. When work is
          present, the poll interval is 1.

        :param sentinel: Breaks the loop
        :type sentinel: Boolean
        """

        self.bind_backend = self.driver.backend_bind()
        poller_time = time.time()
        poller_interval = 128

        while True:
            poller_interval = utils.return_poller_interval(
                poller_time=poller_time,
                poller_interval=poller_interval,
                log=self.log,
            )

            if self.driver.bind_check(
                bind=self.bind_backend, constant=poller_interval
            ):
                poller_interval, poller_time = 128, time.time()
                (
                    identity,
                    msg_id,
                    control,
                    command,
                    data,
                    info,
                    stderr,
                    stdout,
                ) = self.driver.socket_recv(socket=self.bind_backend)
                if control in [
                    self.driver.coordination_notice,
                    self.driver.coordination_ack,
                    self.driver.coordination_failed,
                ]:
                    for _ in range(120):
                        try:
                            self.driver.socket_send(
                                socket=self.bind_backend,
                                identity=info,
                                control=control,
                                command=command,
                                data=data,
                                info=identity,
                                stderr=stderr,
                                stdout=stdout,
                            )
                            break
                        except Exception as e:
                            self.log.debug(
                                "Job [ %s ] connecting to target [ %s ] saw"
                                " exception %s -- retrying",
                                msg_id.decode(),
                                info.decode(),
                                str(e),
                            )
                            time.sleep(0.01)
                    else:
                        try:
                            self.driver.socket_send(
                                socket=self.bind_backend,
                                identity=identity,
                                control=self.driver.coordination_failed,
                                command=command,
                                data=data,
                                info=info,
                                stderr=(
                                    "Failed to connect to coordination node"
                                    " [ {} ] after three attempts.".format(
                                        info.decode()
                                    )
                                ).encode(),
                                stdout=stdout,
                            )
                        except Exception as e:
                            self.log.error(
                                "Job [ %s ] connecting to target [ %s ] saw"
                                " exception %s",
                                msg_id.decode(),
                                info.decode(),
                                str(e),
                            )
                elif control == self.driver.transfer_start:
                    transfer_identity = identity.decode()
                    transfer_job_id = msg_id.decode()
                    transfer_file_path = os.path.abspath(
                        os.path.expanduser(info.decode())
                    )
                    offset = int(command)
                    chunk_size = int(data)
                    self.log.debug(
                        "Identity [ %s ] transfer job [ %s ] processing"
                        " file [ %s ]",
                        transfer_identity,
                        transfer_job_id,
                        transfer_file_path,
                    )
                    if not os.path.isfile(transfer_file_path):
                        self.log.error(
                            "Identity [ %s ] Job [ %s ] File was not found."
                            " File path [ %s ]",
                            transfer_identity,
                            transfer_job_id,
                            transfer_file_path,
                        )
                        self.driver.socket_send(
                            socket=self.bind_backend,
                            identity=transfer_identity.encode(),
                            control=self.driver.job_failed,
                            info="File [ {} ] was not found".format(
                                transfer_file_path
                            ).encode(),
                        )
                    else:
                        self.log.info(
                            "Identity [ %s ] Job [ %s ] file transfer for"
                            " [ %s ] starting",
                            transfer_identity,
                            transfer_job_id,
                            transfer_file_path,
                        )
                        with open(transfer_file_path, "rb") as f:
                            f.seek(offset, os.SEEK_SET)
                            data = f.read(chunk_size)
                            self.driver.socket_send(
                                socket=self.bind_backend,
                                identity=transfer_identity.encode(),
                                control=(
                                    self.driver.transfer_end
                                    if len(data) < chunk_size
                                    else self.driver.job_processing
                                ),
                                data=data,
                            )
                        self.log.info(
                            "Identity [ %s ] Job [ %s ] file transfer for"
                            " [ %s ] blob sent",
                            transfer_identity,
                            transfer_job_id,
                            transfer_file_path,
                        )
                else:
                    self.log.warning(
                        "Unknown transfer job [ %s ] connection received from"
                        " [ %s ], control [ %s ] with"
                        " info [ %s ]",
                        msg_id.decode(),
                        identity.decode(),
                        control.decode(),
                        info.decode(),
                    )

            if sentinel:
                break

    def run_interactions(self, sentinel=False):
        """Execute the interactions loop.

        Directord's interaction executor will slow down the poll interval
        when no work is present. This means Directord will ramp-up resource
        utilization when required and become virtually idle when there's
        nothing to do.

        * Initial poll interval is 1024, maxing out at 2048. When work is
          present, the poll interval is 1.

        :param sentinel: Breaks the loop
        :type sentinel: Boolean
        """

        self.bind_job = self.driver.job_bind()
        poller_time = time.time()
        prune_time = time.time()
        poller_interval = 1
        run_jobs_thread = None
        while True:
            if self.terminate_process(process=run_jobs_thread):
                run_jobs_thread = None

            if not self.job_queue.empty():
                if not run_jobs_thread:
                    run_jobs_thread = self.thread(
                        target=self.run_job,
                        name="run_job",
                        daemon=True,
                    )
                    run_jobs_thread.start()
            elif self.send_queue.empty():
                poller_interval = utils.return_poller_interval(
                    poller_time=poller_time,
                    poller_interval=poller_interval,
                    log=self.log,
                )

            while not self.send_queue.empty():
                try:
                    send_item = self.send_queue.get_nowait()
                except Exception:
                    break
                else:
                    self.log.debug(
                        "Sending job [ %s ] sent to [ %s ]",
                        send_item["data"]["job_id"],
                        send_item["identity"].decode(),
                    )
                    send_item["data"] = json.dumps(send_item["data"]).encode()
                    self.driver.socket_send(
                        socket=self.bind_job,
                        **send_item,
                    )

            if self.driver.bind_check(
                bind=self.bind_job, constant=poller_interval
            ):
                (
                    identity,
                    msg_id,
                    control,
                    _,
                    data,
                    info,
                    stderr,
                    stdout,
                ) = self.driver.socket_recv(socket=self.bind_job)
                if control == self.driver.heartbeat_notice:
                    self.log.debug(
                        "Received Heartbeat from [ %s ]",
                        identity.decode(),
                    )
                    expire = self.driver.get_expiry(
                        heartbeat_interval=self.heartbeat_interval,
                    )
                    metadata = {"time": expire}
                    if data:
                        try:
                            loaded_data = json.loads(data.decode())
                        except Exception:
                            pass
                        else:
                            if loaded_data:
                                metadata.update(loaded_data)

                    self.workers[identity.decode()] = metadata
                else:
                    poller_interval, poller_time = 1, time.time()
                    self.log.debug(
                        "Execution job received [ %s ]", msg_id.decode()
                    )
                    node = identity.decode()
                    node_output = info.decode()
                    if stderr:
                        stderr = stderr.decode()
                    if stdout:
                        stdout = stdout.decode()

                    try:
                        data_item = json.loads(data.decode())
                    except Exception:
                        data_item = dict()

                    self._set_job_status(
                        job_status=control,
                        job_id=msg_id.decode(),
                        identity=node,
                        job_output=node_output,
                        job_stdout=stdout,
                        job_stderr=stderr,
                        execution_time=data_item.get("execution_time", 0),
                        return_timestamp=data_item.get("return_timestamp", 0),
                        component_exec_timestamp=data_item.get(
                            "component_exec_timestamp", 0
                        ),
                        recv_time=time.time(),
                    )

                    for new_task in data_item.get("new_tasks", list()):
                        self.log.debug("New task found: %s", new_task)
                        if "targets" in new_task and new_task["targets"]:
                            targets = [i.encode() for i in new_task["targets"]]
                            self.log.debug(
                                "Using existing targets from old job"
                                " specification %s",
                                targets,
                            )
                        else:
                            targets = self.workers.keys()
                            self.log.debug(
                                "Targets undefined in old job specification"
                                " running everwhere"
                            )

                        # NOTE(cloudnull): If the new task identity is set but
                        #                  with a null value, reset the value
                        #                  to that of the known workers.
                        if "identity" in new_task and not new_task["identity"]:
                            self.log.debug("identities reset to all workers")
                            new_task["identity"] = [
                                i.decode() for i in self.workers.keys()
                            ]

                        if "job_id" not in new_task:
                            new_task["job_id"] = utils.get_uuid()

                        self.create_return_jobs(
                            task=new_task["job_id"],
                            job_item=new_task,
                            targets=targets,
                        )

                        for target in targets:
                            self.log.debug(
                                "Queuing callback job [ %s ] for identity"
                                " [ %s ]",
                                new_task["job_id"],
                                target.decode(),
                            )
                            self.send_queue.put(
                                dict(
                                    identity=target,
                                    command=new_task["verb"].encode(),
                                    data=new_task,
                                )
                            )

            if time.time() > prune_time:
                self.workers.prune()
                prune_time = time.time() + 762

            if sentinel:
                break

    def run_socket_server(self, sentinel=False):
        """Start a socket server.

        The socket server is used to broker a connection from the end user
        into the directord sub-system. The socket server will allow for 1
        message of 10M before requiring the client to reconnect.

        All received data is expected to be JSON serialized data. Before
        being added to the queue, a task ID and SHA3_224 SUM is added to the
        content. This is done for tracking and caching purposes. The task
        ID can be defined in the data. If a task ID is not defined one will
        be generated.

        :param sentinel: Breaks the loop
        :type sentinel: Boolean
        """

        try:
            os.unlink(self.args.socket_path)
        except OSError:
            if os.path.exists(self.args.socket_path):
                raise SystemExit(
                    "Socket path already exists and wasn't able to be"
                    " cleaned up: {}".format(self.args.socket_path)
                )

        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.bind(self.args.socket_path)
        self.log.debug("Socket:%s bound", self.args.socket_path)
        os.chmod(self.args.socket_path, 509)
        uid = 0
        group = getattr(self.args, "socket_group", "root")
        try:
            gid = int(group)
        except ValueError:
            gid = grp.getgrnam(group).gr_gid
        os.chown(self.args.socket_path, uid, gid)
        sock.listen(1)
        while True:
            conn, _ = sock.accept()
            with conn:
                data = conn.recv(409600)
                data_decoded = data.decode()
                json_data = json.loads(data_decoded)
                if "manage" in json_data:
                    self.log.debug("Received manage command: %s", json_data)
                    key, value = next(iter(json_data["manage"].items()))
                    if key == "list_nodes":
                        data = list()
                        for k, v in self.workers.items():
                            if v:
                                expiry = v.pop("time") - time.time()
                                v["expiry"] = expiry
                                try:
                                    data.append((k.decode(), v))
                                except AttributeError:
                                    data.append((str(k), v))
                    elif key == "list_jobs":
                        data = [
                            (str(k), v) for k, v in self.return_jobs.items()
                        ]
                    elif key == "job_info":
                        try:
                            data = [(str(value), self.return_jobs[value])]
                        except KeyError:
                            data = []
                    elif key == "purge_nodes":
                        self.workers.empty()
                        data = {"success": True}
                    elif key == "purge_jobs":
                        self.return_jobs.empty()
                        data = {"success": True}
                    else:
                        data = {"failed": True}

                    try:
                        conn.sendall(json.dumps(data).encode())
                    except BrokenPipeError as e:
                        self.log.error(
                            "Encountered a broken pipe while sending manage"
                            " data. Error:%s",
                            str(e),
                        )
                else:
                    json_data["job_id"] = json_data.get(
                        "job_id", utils.get_uuid()
                    )

                    if "parent_id" not in json_data:
                        json_data["parent_id"] = json_data["job_id"]

                    # Returns the message in reverse to show a return. This
                    # will be a standard client return in JSON format under
                    # normal circomstances.
                    if json_data.get("return_raw", False):
                        msg = json_data["job_id"].encode()
                    else:
                        msg = "Job received. Task ID: {}".format(
                            json_data["job_id"]
                        ).encode()

                    try:
                        conn.sendall(msg)
                    except BrokenPipeError as e:
                        self.log.error(
                            "Encountered a broken pipe while sending job"
                            " data. Error:%s",
                            str(e),
                        )
                    else:
                        self.log.debug("Data sent to queue [ %s ]", json_data)
                        self.job_queue.put(json_data)

            if sentinel:
                break

    def worker_run(self):
        """Run all work related threads.

        Threads are gathered into a list of process objects then fed into the
        run_threads method where their execution will be managed.
        """

        threads = [
            (
                self.thread(
                    name="run_socket_server", target=self.run_socket_server
                ),
                True,
            ),
            (
                self.thread(
                    name="run_interactions", target=self.run_interactions
                ),
                False,
            ),
            (
                self.thread(name="run_backend", target=self.run_backend),
                True,
            ),
        ]

        if self.args.run_ui:
            # low import to ensure nothing flask is loading needlessly.
            from directord import ui  # noqa

            ui_obj = ui.UI(
                args=self.args, jobs=self.return_jobs, nodes=self.workers
            )
            threads.append(
                (self.thread(name="ui", target=ui_obj.start_app), True)
            )

        self.run_threads(threads=threads)



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import ast
import json

from directord import components


class Component(components.ComponentBase):
    command = "arg"

    def __init__(self):
        """Initialize the component cache class.

        This component is not cacheable.
        """

        super().__init__(desc="Process cache commands")
        self.cacheable = False
        self.requires_lock = True

    def args(self, cache_type):
        """Set default arguments for a component."""

        super().args()
        self.parser.add_argument(
            "--extend-args",
            action="store_true",
            help=(
                "When setting an ARG, allow complex args to extend"
                " existing ones."
            ),
        )
        self.parser.add_argument(
            cache_type,
            nargs="+",
            action="append",
            help="Set a given argument. KEY VALUE",
        )

    def server(self, exec_array, data, arg_vars):
        """Return data from formatted transfer action.

        :param exec_array: Input string from action
        :type exec_array: List
        :param data: Formatted data hash
        :type data: Dictionary
        :param arg_vars: Pre-Formatted arguments
        :type arg_vars: Dictionary
        :returns: Dictionary
        """

        cache_type = "{}s".format(self.verb.lower())
        self.args(cache_type=cache_type)
        args, _ = self.exec_parser(
            parser=self.parser, exec_array=exec_array, arg_vars=arg_vars
        )
        cache_obj = getattr(args, cache_type)
        key, value = " ".join(cache_obj[0]).split(" ", 1)
        try:
            value = ast.literal_eval(value)
        except (ValueError, SyntaxError):
            pass

        if args.extend_args:
            data["extend_args"] = args.extend_args

        data[cache_type] = {key: value}
        return data

    def client(self, cache, job):
        """Run cache command operation.

        :param cache: Caching object used to template items within a command.
        :type cache: Object
        :param job: Information containing the original job specification.
        :type job: Dictionary
        :param command: Work directory path.
        :type command: String
        :returns: tuple
        """

        self.log.debug("client(): job: %s, cache: %s", job, cache)
        # Sets the cache type to "args" or "envs"
        cache_type = "{}s".format(self.command.decode().lower())

        try:
            cache_value = ast.literal_eval(job[cache_type])
        except (ValueError, SyntaxError):
            cache_value = job[cache_type]

        self.log.debug("Job [ %s ] ARG value: %s", job["job_id"], cache_value)
        success, value = self.blueprinter(
            content=json.dumps(cache_value),
            values=cache.get("args"),
            allow_empty_values=True,
        )
        if success:
            cache_value = json.loads(value)

        if cache_value:
            cache_set = self.set_cache(
                cache=cache,
                key=cache_type,
                value=cache_value,
                value_update=True,
                tag=cache_type,
                extend=job.get("extend_args", False),
            )
            if cache_set:
                self.log.debug(
                    "ARG [ %s ] = [ %s ] added to cache",
                    cache_type,
                    cache_value,
                )
                return (
                    "{} added to cache".format(cache_type),
                    None,
                    True,
                    "type:{}, value:{}".format(
                        cache_type, cache_value
                    ).encode(),
                )
            else:
                return (
                    None,
                    "Failed to add {} to cache".format(cache_type),
                    False,
                    "type:{}, value:{}".format(
                        cache_type, cache_value
                    ).encode(),
                )
        else:
            return (
                "Nothing added to cache. {} had no value".format(cache_type),
                None,
                True,
                "type:{}, value:{}".format(cache_type, cache_value).encode(),
            )



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
import glob
import os
import setuptools

from directord import meta


with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()


REQUIREMENTS = {
    "ui": ["flask"],
    "dev": ["podman-py"],
    "test": ["flake8", "coverage"],
    "redis": ["redis"],
}
REQUIREMENTS["all"] = [item for line in REQUIREMENTS.values() for item in line]


setuptools.setup(
    name="directord",
    author=meta.__author__,
    author_email=meta.__email__,
    description=(
        "A deployment framework built to manage the data center life cycle."
    ),
    version=meta.__version__,
    packages=["directord"],
    include_package_data=True,
    zip_safe=False,
    test_suite="tests",
    install_requires=[
        "diskcache",
        "jinja2",
        "pyyaml",
        "pyzmq",
        "ssh-python",
        "tabulate",
        "tenacity",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cloudnull/directord",
    project_urls={
        "Bug Tracker": "https://github.com/cloudnull/directord/issues",
    },
    python_requires=">=3.6",
    extras_require=REQUIREMENTS,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Topic :: Utilities",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    entry_points={
        "console_scripts": [
            "directord = directord.main:main",
            "directord-server-systemd = directord.main:_systemd_server",
            "directord-client-systemd = directord.main:_systemd_client",
        ]
    },
    data_files=[
        (
            "share/directord/components",
            [i for i in glob.glob("components/*") if os.path.isfile(i)],
        ),
        (
            "share/directord/orchestrations",
            [i for i in glob.glob("orchestrations/*") if os.path.isfile(i)],
        ),
        (
            "share/directord/orchestrations/files",
            [
                i
                for i in glob.glob("orchestrations/files/*")
                if os.path.isfile(i)
            ],
        ),
        (
            "share/directord/pods",
            [i for i in glob.glob("pods/*") if os.path.isfile(i)],
        ),
        (
            "share/directord/tools",
            [i for i in glob.glob("tools/*") if os.path.isfile(i)],
        ),
    ],
)



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import argparse
import os
import subprocess

from multiprocessing import Event

import jinja2
from jinja2 import StrictUndefined

import yaml

from directord import logger
from directord import utils


class ComponentBase:
    """Component base class."""

    command = None
    info = None
    driver = None
    verb = None
    block_on_tasks = None
    queue_sentinel = False
    delay = Event().wait

    def __init__(self, desc=None):
        """Initialize the component base class.

        When setting up a component, the init should be inheritted allowing
        user defined components to have access to the full suite of defaults.

        > Set the `self.cacheable` object True|False according to how the
          component should be treated in terms of on system cache.
        """

        self.desc = desc
        self.log = logger.getLogger(name="directord")
        self.blueprint = jinja2.Environment(
            loader=jinja2.BaseLoader(),
            keep_trailing_newline=True,
            undefined=StrictUndefined,
        )
        self.known_args = None
        self.unknown_args = None
        self.cacheable = True  # Enables|Disables component caching
        self.requires_lock = False  # Enables|Disables component locking

    @staticmethod
    def run_command(
        command,
        shell=True,
        env=None,
        execute="/bin/sh",
        return_codes=None,
        no_block=False,
    ):
        """Run a shell command.

        The options available:

        * `shell` to be enabled or disabled, which provides the ability
        to execute arbitrary stings or not. if disabled commands must be
        in the format of a `list`

        * `env` is an environment override and or manipulation setting
        which sets environment variables within the locally executed
        shell.

        * `execute` changes the interpreter which is executing the
        command(s).

        * `return_codes` defines the return code that the command must
        have in order to ensure success. This can be a list of return
        codes if multiple return codes are acceptable.

        :param command: String
        :param shell: Boolean
        :param env: Dictionary
        :param execute: String
        :param return_codes: Integer|List
        :param no_block: Boolean
        :returns: Truple
        """

        if env:
            _env = dict(os.environ)
            _env.update(env)
            env = _env
        else:
            env = os.environ

        stdout = subprocess.PIPE

        if return_codes is None:
            return_codes = [0]
        if isinstance(return_codes, int):
            return_codes = [return_codes]

        stderr = subprocess.PIPE
        process = subprocess.Popen(
            command,
            stdout=stdout,
            stderr=stderr,
            executable=execute,
            env=env,
            shell=shell,
            start_new_session=no_block,
        )
        if no_block:
            return None, None, True

        output, error = process.communicate()
        if process.returncode not in return_codes:
            return output, error, False

        return output, error, True

    def options_converter(self, documentation):
        """Convert an options YAML to Arguments.

        :param documentation: YAML content.
        :type documentation: String
        """

        argument_spec = yaml.safe_load(documentation)["options"]
        for key, value in argument_spec.items():
            options = dict()
            description = value.pop("description", None)
            if description:
                if isinstance(description, list):
                    description = " ".join(description)
                options["help"] = description

            default = value.pop("default", None)
            if default:
                options["default"] = default

            required = value.pop("required", None)
            if isinstance(required, bool):
                options["required"] = required
            elif isinstance(required, str) and required.lower() in [
                "yes",
                "true",
            ]:
                options["required"] = True

            arg_type = value.pop("type", None)
            if isinstance(arg_type, str):
                arg_type = arg_type.lower()
                if arg_type == "bool":
                    options["action"] = "store_true"
                elif arg_type == "list":
                    options["type"] = list
                elif arg_type == "dict":
                    options["type"] = dict
                elif arg_type == "int":
                    options["type"] = int
                elif arg_type == "str":
                    options["type"] = str

            choices = value.pop("choices", None)
            if choices:
                options["choices"] = choices

            self.parser.add_argument(
                "--{}".format(key.replace("_", "-")), **options
            )

    @staticmethod
    def sanitized_args(execute):
        """Return arguments in a flattened array.

        This will inspect the execution arguments and return everything found
        as a flattened array.

        :param execute: Execution string to parse.
        :type execute: String
        :returns: List
        """

        if execute:
            return [i for g in execute if g for i in g.split()]
        else:
            return list()

    def exec_parser(self, parser, exec_array, arg_vars=None):
        """Run the parser and return parsed arguments.

        :param parser: Argument parser.
        :type parser: Object
        :param exec_array: Input array from action
        :type exec_array: List
        :param arg_vars: Pre-Formatted arguments
        :type arg_vars: Dictionary
        :returns: Tuple
        """

        self.known_args, self.unknown_args = parser.parse_known_args(
            self.sanitized_args(execute=exec_array)
        )
        if hasattr(self.known_args, "exec_help") and self.known_args.exec_help:
            raise SystemExit(parser.print_help())
        else:
            if arg_vars:
                for key, value in arg_vars.items():
                    self.known_args.__dict__[key] = value
            return self.known_args, self.unknown_args

    def args(self):
        """Set default arguments for a component."""

        self.parser = argparse.ArgumentParser(
            description=self.desc,
            allow_abbrev=False,
            add_help=False,
        )
        self.parser.add_argument(
            "--exec-help",
            action="help",
            help="Show this execution help message.",
        )
        self.parser.add_argument(
            "--skip-cache",
            action="store_true",
            help="For a task to skip the on client cache.",
        )
        self.parser.add_argument(
            "--run-once",
            action="store_true",
            help="Force a given task to run once.",
        )
        self.parser.add_argument(
            "--timeout",
            default=600,
            type=int,
            help="Set the action timeout. Default %(default)s.",
        )
        self.parser.add_argument(
            "--force-lock",
            action="store_true",
            help="Force a given task to run with a lock.",
        )

    @staticmethod
    def set_cache(
        cache,
        key,
        value,
        value_update=False,
        expire=259200,
        tag=None,
        extend=False,
    ):
        """Set a cached item.

        :param cache: Cached access object.
        :type cache: Object
        :param key: Key for the cached item.
        :type key: String
        :param value: Value for the cached item.
        :type value: ANY
        :param value_update: Instructs the method to update a Dictionary with
                             another dictionary.
        :type value_update: Boolean
        :param expire: Sets the expire time, defaults to 12 hours.
        :type expire: Integer
        :param tag: Sets the index for a given cached item.
        :type tag: String
        :param extend: Enable|Disable Extend a map
        :type extend: Boolean
        :returns" Boolean
        """

        if value_update:
            orig = cache.get(key, default=dict())
            value = utils.merge_dict(orig, value, extend=extend)
        else:
            try:
                value = value.decode()
            except (ValueError, AttributeError):
                pass

        cache_set = cache.set(key, value, tag=tag, expire=expire, retry=True)
        if not cache_set:
            return cache.set(key, value, tag=tag, expire=expire, retry=True)
        return cache_set

    def file_blueprinter(self, cache, file_to):
        """Read a file and blueprint its contents.

        :param cache: Cached access object.
        :type cache: Object
        :param file_to: String path to a file which will blueprint.
        :type file_to: String
        :returns: Boolean
        """

        try:
            with open(file_to) as f:
                success, contents = self.blueprinter(
                    content=f.read(), values=cache.get("args")
                )

            if not success:
                return False, contents

            with open(file_to, "w") as f:
                f.write(contents)
        except Exception as e:
            error = str(e)
            self.log.critical("File blueprint failure: %s", error)
            return False, error
        else:
            self.log.info("File %s has been blueprinted.", file_to)
            return True, None

    def blueprinter(self, content, values, allow_empty_values=False):
        """Return blue printed content.

        :param content: A string item that will be interpreted and blueprinted.
        :type content: String
        :param values: Dictionary items that will be used to render a
                       blueprinted item.
        :type values: Dictionary
        :param allow_empty_values: Enable|Disable null values
        :type allow_empty_values: Boolean
        :returns: Tuple
        """

        if not values:
            if allow_empty_values:
                values = dict()
            else:
                return False, "No arguments were defined for blueprinting"

        try:
            _contents = self.blueprint.from_string(content)
            rendered_content = _contents.render(**values)
        except Exception as e:
            error = str(e)
            self.log.debug("blueprint error: %s, values: %s", error, values)
            return False, error
        else:
            return True, rendered_content

    def parser_error(self):
        """Return parser help information."""

        return self.parser.print_help()

    def server(self, exec_array, data, arg_vars):
        """Server operation.

        :param exec_array: Input array from action
        :type exec_array: List
        :param data: Formatted data hash
        :type data: Dictionary
        :returns: Dictionary
        """

        self.args()
        self.exec_parser(
            parser=self.parser, exec_array=exec_array, arg_vars=arg_vars
        )

    def client(self, cache, job):
        """Client operation.

        Command operations are rendered with cached data from the args dict.

        :param cache: Caching object used to template items within a command.
        :type cache: Object
        :param job: Information containing the original job specification.
        :type job: Dictionary
        """

        pass

    def close(self):
        """Close a component."""

        pass

    def __enter__(self):
        """Return self when using a component as a context manager."""

        return self

    def __exit__(self, *args, **kwargs):
        """Exit the context manager and close."""

        self.close()



#   Copyright Peznauts <kevin@cloudnull.com>. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import logging
import os

import tenacity
import zmq
import zmq.auth as zmq_auth
from zmq.auth.thread import ThreadAuthenticator

from directord import drivers
from directord import logger
from directord import utils


class Driver(drivers.BaseDriver):
    def __init__(
        self, args, encrypted_traffic_data=None, connection_string=None
    ):
        """Initialize the Driver.

        :param args: Arguments parsed by argparse.
        :type args: Object
        :param encrypted_traffic: Enable|Disable encrypted traffic.
        :type encrypted_traffic: Boolean
        :param connection_string: Connection string used to provide connection
                                  instructions to the driver.
        :type connection_string: String.
        """

        self.args = args
        self.encrypted_traffic_data = encrypted_traffic_data
        self.connection_string = connection_string
        if self.encrypted_traffic_data:
            self.encrypted_traffic = self.encrypted_traffic_data.get("enabled")
            self.secret_keys_dir = self.encrypted_traffic_data.get(
                "secret_keys_dir"
            )
            self.public_keys_dir = self.encrypted_traffic_data.get(
                "public_keys_dir"
            )
        else:
            self.encrypted_traffic = False
            self.secret_keys_dir = None
            self.public_keys_dir = None

        self.ctx = zmq.Context().instance()
        self.poller = zmq.Poller()
        super(Driver, self).__init__(
            args=args,
            encrypted_traffic_data=self.encrypted_traffic_data,
            connection_string=self.connection_string,
        )

    def __copy__(self):
        return Driver(
            args=self.args,
            encrypted_traffic_data=self.encrypted_traffic_data,
            connection_string=self.connection_string,
        )

    def _socket_context(self, socket_type):
        """Create socket context and return a bind object.

        :param socket_type: Set the Socket type, typically defined using a ZMQ
                            constant.
        :type socket_type: Integer
        :returns: Object
        """

        bind = self.ctx.socket(socket_type)
        # NOTE(cloudnull): STUPID SOLUTION. We should have
        #                  a more intelligent HWM solution.
        try:
            bind.sndhwm = bind.rcvhwm = 0
        except AttributeError:
            bind.hwm = 0

        bind.set_hwm(0)
        bind.setsockopt(zmq.SNDHWM, 0)
        bind.setsockopt(zmq.RCVHWM, 0)
        if socket_type == zmq.ROUTER:
            bind.setsockopt(zmq.ROUTER_MANDATORY, 1)
        return bind

    def _socket_bind(
        self, socket_type, connection, port, poller_type=zmq.POLLIN
    ):
        """Return a socket object which has been bound to a given address.

        When the socket_type is not PUB or PUSH, the bound socket will also be
        registered with self.poller as defined within the Interface
        class.

        :param socket_type: Set the Socket type, typically defined using a ZMQ
                            constant.
        :type socket_type: Integer
        :param connection: Set the Address information used for the bound
                           socket.
        :type connection: String
        :param port: Define the port which the socket will be bound to.
        :type port: Integer
        :param poller_type: Set the Socket type, typically defined using a ZMQ
                            constant.
        :type poller_type: Integer
        :returns: Object
        """

        bind = self._socket_context(socket_type=socket_type)
        auth_enabled = self.args.shared_key or self.args.curve_encryption

        if auth_enabled:
            self.auth = ThreadAuthenticator(self.ctx, log=self.log)
            self.auth.start()
            self.auth.allow()

            if self.args.shared_key:
                # Enables basic auth
                self.auth.configure_plain(
                    domain="*", passwords={"admin": self.args.shared_key}
                )
                bind.plain_server = True  # Enable shared key authentication
                self.log.info("Shared key authentication enabled.")
            elif self.args.curve_encryption:
                server_secret_file = os.path.join(
                    self.secret_keys_dir, "server.key_secret"
                )
                for item in [
                    self.public_keys_dir,
                    self.secret_keys_dir,
                    server_secret_file,
                ]:
                    if not os.path.exists(item):
                        raise SystemExit(
                            "The required path [ {} ] does not exist. Have"
                            " you generated your keys?".format(item)
                        )
                self.auth.configure_curve(
                    domain="*", location=self.public_keys_dir
                )
                try:
                    server_public, server_secret = zmq_auth.load_certificate(
                        server_secret_file
                    )
                except OSError as e:
                    self.log.error(
                        "Failed to load certificates: %s, Configuration: %s",
                        str(e),
                        vars(self.args),
                    )
                    raise SystemExit("Failed to load certificates")
                else:
                    bind.curve_secretkey = server_secret
                    bind.curve_publickey = server_public
                    bind.curve_server = True  # Enable curve authentication
        bind.bind(
            "{connection}:{port}".format(
                connection=connection,
                port=port,
            )
        )

        if socket_type not in [zmq.PUB]:
            self.poller.register(bind, poller_type)

        return bind

    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(TimeoutError),
        wait=tenacity.wait_fixed(5),
        before_sleep=tenacity.before_sleep_log(
            logger.getLogger(name="directord"), logging.WARN
        ),
    )
    def _socket_connect(
        self, socket_type, connection, port, poller_type=zmq.POLLIN
    ):
        """Return a socket object which has been bound to a given address.

        > A connection back to the server will wait 10 seconds for an ack
          before going into a retry loop. This is done to forcefully cycle
          the connection object to reset.

        :param socket_type: Set the Socket type, typically defined using a ZMQ
                            constant.
        :type socket_type: Integer
        :param connection: Set the Address information used for the bound
                           socket.
        :type connection: String
        :param port: Define the port which the socket will be bound to.
        :type port: Integer
        :param poller_type: Set the Socket type, typically defined using a ZMQ
                            constant.
        :type poller_type: Integer
        :returns: Object
        """

        bind = self._socket_context(socket_type=socket_type)

        if self.args.shared_key:
            bind.plain_username = b"admin"  # User is hard coded.
            bind.plain_password = self.args.shared_key.encode()
            self.log.info("Shared key authentication enabled.")
        elif self.args.curve_encryption:
            client_secret_file = os.path.join(
                self.secret_keys_dir, "client.key_secret"
            )
            server_public_file = os.path.join(
                self.public_keys_dir, "server.key"
            )
            for item in [
                self.public_keys_dir,
                self.secret_keys_dir,
                client_secret_file,
                server_public_file,
            ]:
                if not os.path.exists(item):
                    raise SystemExit(
                        "The required path [ {} ] does not exist. Have"
                        " you generated your keys?".format(item)
                    )
            try:
                client_public, client_secret = zmq_auth.load_certificate(
                    client_secret_file
                )
                server_public, _ = zmq_auth.load_certificate(
                    server_public_file
                )
            except OSError as e:
                self.log.error(
                    "Error while loading certificates: %s. Configuration: %s",
                    str(e),
                    vars(self.args),
                )
                raise SystemExit("Failed to load keys.")
            else:
                bind.curve_secretkey = client_secret
                bind.curve_publickey = client_public
                bind.curve_serverkey = server_public

        if socket_type == zmq.SUB:
            bind.setsockopt_string(zmq.SUBSCRIBE, self.identity)
        else:
            bind.setsockopt_string(zmq.IDENTITY, self.identity)

        self.poller.register(bind, poller_type)
        bind.connect(
            "{connection}:{port}".format(
                connection=connection,
                port=port,
            )
        )

        self.log.info("Socket connected to [ %s ].", connection)
        return bind

    def socket_send(
        self,
        socket,
        identity=None,
        msg_id=None,
        control=None,
        command=None,
        data=None,
        info=None,
        stderr=None,
        stdout=None,
        nonblocking=False,
    ):
        """Send a message over a ZM0 socket.

        The message specification for server is as follows.

            [
                b"Identity"
                b"ID",
                b"ASCII Control Characters",
                b"command",
                b"data",
                b"info",
                b"stderr",
                b"stdout",
            ]

        The message specification for client is as follows.

            [
                b"ID",
                b"ASCII Control Characters",
                b"command",
                b"data",
                b"info",
                b"stderr",
                b"stdout",
            ]

        All message information is assumed to be byte encoded.

        All possible control characters are defined within the Interface class.
        For more on control characters review the following
        URL(https://donsnotes.com/tech/charsets/ascii.html#cntrl).

        :param socket: ZeroMQ socket object.
        :type socket: Object
        :param identity: Target where message will be sent.
        :type identity: Bytes
        :param msg_id: ID information for a given message. If no ID is
                       provided a UUID will be generated.
        :type msg_id: Bytes
        :param control: ASCII control charaters.
        :type control: Bytes
        :param command: Command definition for a given message.
        :type command: Bytes
        :param data: Encoded data that will be transmitted.
        :type data: Bytes
        :param info: Encoded information that will be transmitted.
        :type info: Bytes
        :param stderr: Encoded error information from a command.
        :type stderr: Bytes
        :param stdout: Encoded output information from a command.
        :type stdout: Bytes
        :param nonblocking: Enable non-blocking send.
        :type nonblocking: Boolean
        """

        if not msg_id:
            msg_id = utils.get_uuid().encode()

        if not control:
            control = self.nullbyte

        if not command:
            command = self.nullbyte

        if not data:
            data = self.nullbyte

        if not info:
            info = self.nullbyte

        if not stderr:
            stderr = self.nullbyte

        if not stdout:
            stdout = self.nullbyte

        message_parts = [msg_id, control, command, data, info, stderr, stdout]

        if identity:
            message_parts.insert(0, identity)

        if nonblocking:
            flags = zmq.NOBLOCK
        else:
            flags = 0

        return socket.send_multipart(message_parts, flags=flags)

    @staticmethod
    def socket_recv(socket, nonblocking=False):
        """Receive a message over a ZM0 socket.

        The message specification for server is as follows.

            [
                b"Identity"
                b"ID",
                b"ASCII Control Characters",
                b"command",
                b"data",
                b"info",
                b"stderr",
                b"stdout",
            ]

        The message specification for client is as follows.

            [
                b"ID",
                b"ASCII Control Characters",
                b"command",
                b"data",
                b"info",
                b"stderr",
                b"stdout",
            ]

        All message parts are byte encoded.

        All possible control characters are defined within the Interface class.
        For more on control characters review the following
        URL(https://donsnotes.com/tech/charsets/ascii.html#cntrl).

        :param socket: ZeroMQ socket object.
        :type socket: Object
        :param nonblocking: Enable non-blocking receve.
        :type nonblocking: Boolean
        """

        if nonblocking:
            flags = zmq.NOBLOCK
        else:
            flags = 0

        return socket.recv_multipart(flags=flags)

    def job_connect(self):
        """Connect to a job socket and return the socket.

        :returns: Object
        """

        self.log.debug("Establishing Job connection.")
        return self._socket_connect(
            socket_type=zmq.DEALER,
            connection=self.connection_string,
            port=self.args.job_port,
        )

    def backend_connect(self):
        """Connect to a backend socket and return the socket.

        :returns: Object
        """

        self.log.debug("Establishing backend connection.")
        bind = self._socket_connect(
            socket_type=zmq.DEALER,
            connection=self.connection_string,
            port=self.args.backend_port,
        )
        bind.set_hwm(16)
        self.log.debug(
            "Identity [ %s ] backend connect hwm state [ %s ]",
            self.identity,
            bind.get_hwm(),
        )
        return bind

    def job_bind(self):
        """Bind an address to a job socket and return the socket.

        :returns: Object
        """

        return self._socket_bind(
            socket_type=zmq.ROUTER,
            connection=self.connection_string,
            port=self.args.job_port,
        )

    def backend_bind(self):
        """Bind an address to a backend socket and return the socket.

        :returns: Object
        """

        bind = self._socket_bind(
            socket_type=zmq.ROUTER,
            connection=self.connection_string,
            port=self.args.backend_port,
        )
        bind.set_hwm(16)
        self.log.debug(
            "Identity [ %s ] backend connect hwm state [ %s ]",
            self.identity,
            bind.get_hwm(),
        )
        return bind

    def bind_check(self, bind, interval=1, constant=1000):
        """Return True if a bind type contains work ready.

        :param bind: A given Socket bind to identify.
        :type bind: Object
        :param interval: Exponential Interval used to determine the polling
                         duration for a given socket.
        :type interval: Integer
        :param constant: Constant time used to poll for new jobs.
        :type constant: Integer
        :returns: Object
        """

        socks = dict(self.poller.poll(interval * constant))
        return socks.get(bind) == zmq.POLLIN

    def key_generate(self, keys_dir, key_type):
        """Generate certificate.

        :param keys_dir: Full Directory path where a given key will be stored.
        :type keys_dir: String
        :param key_type: Key type to be generated.
        :type key_type: String
        """

        zmq_auth.create_certificates(keys_dir, key_type)

    def create_proxy(front, back):
        """Create proxy bind.

        Bind two interfaces into a proxy.

        :param front: Frontend interface.
        :type front: Object
        :param back: Backend interface.
        :type back: Object
        """

        return zmq.proxy(frontend=front, backend=back)



#   Copyright Alex Schultz <aschultz@next-development.com>. All Rights Reserved
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from directord import components


class Component(components.ComponentBase):
    def __init__(self):
        """Initialize the component cache class."""

        super().__init__(desc="Manage packages with dnf")
        self.requires_lock = True

    def args(self):
        """Set default arguments for a component."""

        super().args()
        self.parser.add_argument(
            "--clear-metadata",
            action="store_true",
            help="Clear dnf metadata and make cache before running action.",
        )
        state_group = self.parser.add_mutually_exclusive_group()
        state_group.add_argument(
            "--latest",
            help="Ensure latest package is installed.",
            action="store_true",
        )
        state_group.add_argument(
            "--absent", help="Ensure packages are removed", action="store_true"
        )
        self.parser.add_argument(
            "packages",
            nargs="+",
            help="A space delineated list of packages to manage.",
        )

    def server(self, exec_array, data, arg_vars):
        """Return data from formatted transfer action.

        :param exec_array: Input array from action
        :type exec_array: List
        :param data: Formatted data hash
        :type data: Dictionary
        :param arg_vars: Pre-Formatted arguments
        :type arg_vars: Dictionary
        :returns: Dictionary
        """

        super().server(exec_array=exec_array, data=data, arg_vars=arg_vars)
        if self.known_args.absent:
            data["state"] = "absent"
        elif self.known_args.latest:
            data["state"] = "latest"
        else:
            data["state"] = "present"

        data["clear"] = self.known_args.clear_metadata
        data["packages"] = self.known_args.packages

        return data

    def client(self, cache, job):
        """Run file command operation.

        Command operations are rendered with cached data from the args dict.

        :param cache: Caching object used to template items within a command.
        :type cache: Object
        :param job: Information containing the original job specification.
        :type job: Dictionary
        :returns: tuple
        """

        self.log.debug("client(): job: %s, cache: %s", job, cache)
        state = job.get("state")
        clear = job.get("clear")

        job_stdout = []
        job_stderr = []
        outcome = False
        if clear:
            cmd = "dnf clean all"
            job_stdout.append(b"=== dnf clean ===\n")
            stdout, stderr, outcome = self.run_command(
                command=cmd, env=cache.get("envs")
            )
            job_stdout.append(stdout)
            job_stderr.append(stderr)
            # TODO: check outcome
            cmd = "dnf makecache"
            job_stdout.append(b"=== dnf makecache ===\n")
            stdout, stderr, outcome = self.run_command(
                command=cmd, env=cache.get("envs")
            )
            job_stdout.append(stdout)
            job_stderr.append(stderr)
            # TODO: check outcome

        packages = job.get("packages")

        if not packages:
            return None, None, False, None

        to_remove = []
        to_install = []
        to_update = []
        if state == "absent":
            to_remove = packages
        elif state == "latest":
            for package in packages:
                cmd = "dnf list --installed {}".format(package)
                stdout, stderr, outcome = self.run_command(
                    command=cmd, env=cache.get("envs")
                )
                if outcome:
                    to_update.append(package)
                else:
                    to_install.append(package)
        else:
            to_install = packages

        if to_remove:
            cmd = "dnf -q -y remove {}".format(" ".join(to_remove))
            job_stdout.append(b"=== dnf remove ===\n")
            stdout, stderr, outcome = self.run_command(
                command=cmd, env=cache.get("envs")
            )
            job_stdout.append(stdout)
            job_stderr.append(stderr)

        if to_install:
            cmd = "dnf -q -y install {}".format(" ".join(to_install))
            job_stdout.append(b"=== dnf install ===\n")
            stdout, stderr, outcome = self.run_command(
                command=cmd, env=cache.get("envs")
            )
            job_stdout.append(stdout)
            job_stderr.append(stderr)

        if to_update:
            cmd = "dnf -q -y update {}".format(" ".join(to_update))
            job_stdout.append(b"=== dnf update ===\n")
            stdout, stderr, outcome = self.run_command(
                command=cmd, env=cache.get("envs")
            )
            job_stdout.append(stdout)
            job_stderr.append(stderr)

        return b"".join(job_stdout), b"".join(job_stderr), outcome, None



#   Copyright Alex Schultz <aschultz@next-development.com>. All Rights Reserved
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from directord import components


class Component(components.ComponentBase):
    def __init__(self):
        """Initialize the component service class."""

        super().__init__(desc="Manage services with systemd")
        self.requires_lock = True

    def args(self):
        """Set default arguments for a component."""

        super().args()
        running_group = self.parser.add_mutually_exclusive_group()
        running_group.add_argument(
            "--restarted",
            action="store_true",
            help="Ensure service is restarted",
        )
        running_group.add_argument(
            "--stopped", action="store_true", help="Ensure service is started"
        )
        state_group = self.parser.add_mutually_exclusive_group()
        state_group.add_argument(
            "--enable",
            help="Ensure service is enabled.",
            action="store_true",
        )
        state_group.add_argument(
            "--disable", help="Ensure service is disabled", action="store_true"
        )
        running_group.add_argument(
            "--daemon-reload",
            action="store_true",
            help="Reload the systemd daemon",
        )
        self.parser.add_argument(
            "services",
            nargs="+",
            help="A space delineated list of services to manage.",
        )

    def server(self, exec_array, data, arg_vars):
        """Return data from formatted transfer action.

        :param exec_array: Input array from action
        :type exec_array: List
        :param data: Formatted data hash
        :type data: Dictionary
        :param arg_vars: Pre-Formatted arguments
        :type arg_vars: Dictionary
        :returns: Dictionary
        """

        super().server(exec_array=exec_array, data=data, arg_vars=arg_vars)
        if self.known_args.enable:
            data["state"] = "enable"
        elif self.known_args.disable:
            data["state"] = "disable"

        elif self.known_args.restarted:
            data["running"] = "restart"
        elif self.known_args.stopped:
            data["running"] = "stop"
        else:
            data["running"] = "start"

        data["services"] = self.known_args.services
        data["daemon_reload"] = self.known_args.daemon_reload

        return data

    def client(self, cache, job):
        """Run service command operation.

        Command operations are rendered with cached data from the args dict.

        :param cache: Caching object used to template items within a command.
        :type cache: Object
        :param job: Information containing the original job specification.
        :type job: Dictionary
        :returns: tuple
        """

        self.log.debug("client(): job: %s, cache: %s", job, cache)
        state = job.get("state")
        running = job.get("running", "start")
        services = job.get("services")

        job_stdout = []
        job_stderr = []
        outcome = False
        if not services:
            return None, None, False, None

        if job.get("daemon_reload") is True:
            stdout, stderr, outcome = self.run_command(
                command="systemctl daemon-reload", env=cache.get("envs")
            )
            job_stdout.append(stdout)
            job_stderr.append(stderr)

            # fail if enablement fails
            if not outcome:
                return (
                    b"".join(job_stdout),
                    b"".join(job_stderr),
                    outcome,
                    "Failed to reload the systemd daemon",
                )

        if state:
            cmd = "systemctl {} {}".format(state, " ".join(services))
            stdout, stderr, outcome = self.run_command(
                command=cmd, env=cache.get("envs")
            )
            job_stdout.append(stdout)
            job_stderr.append(stderr)

            # fail if enablement fails
            if not outcome:
                return (
                    b"".join(job_stdout),
                    b"".join(job_stderr),
                    outcome,
                    "Failed to set the systemd state",
                )

        cmd = "systemctl {} {}".format(running, " ".join(services))
        stdout, stderr, outcome = self.run_command(
            command=cmd, env=cache.get("envs")
        )
        job_stdout.append(stdout)
        job_stderr.append(stderr)

        return b"".join(job_stdout), b"".join(job_stderr), outcome, None
